#include "types.h"
#include "xboxkrnl.h"
#include "boot.h"
#include "video.h"
#include "xbox.h"



/*
 * Detect video encoder. Similar to xbox-linux project's detector in kernel.
 */

encoder_type video_get_encoder_type(void)
{
    static encoder_type type = ENCODER_UNKNOWN;
    static int detected = 0;
    DWORD value;

    if (detected)
	return type;

    detected = 1;
    
    if (NT_SUCCESS(HalReadSMBusValue(0x8a, 0x00, 0, &value))) {
	dprintf(" * Conexant video encoder\n");
	type = ENCODER_CONEXANT;
    }
    else if (NT_SUCCESS(HalReadSMBusValue(0xd4, 0x00, 0, &value))) {
	dprintf(" * Focus video encoder\n");
	type = ENCODER_FOCUS;
    }
    else if (NT_SUCCESS(HalReadSMBusValue(0xe0, 0x01, 0, &value))) {
	dprintf(" * Xcalibur video encoder\n");
	type = ENCODER_XCALIBUR;
    }
    else {
	dprintf(" * Unknown video encoder\n");
	type = ENCODER_UNKNOWN;
    }

    return type;
}



void video_off(void)
{
    switch (video_get_encoder_type()) {
    case ENCODER_CONEXANT:
        IoOutputByte(0x80d3, 0x05);
	HalWriteSMBusValue(0x8a, 0xba, 0, 0x3f);
	break;

    case ENCODER_FOCUS:
	HalWriteSMBusValue(0xd4, 0xa0, 1, 0x020f);
	break;

    case ENCODER_XCALIBUR:
	HalWriteSMBusValue(0xe0, 0x04, 1, 0x000f);
	break;

    case ENCODER_UNKNOWN:
    default:
	break;
    }
}



/*
 * Move current framebuffer to a location where it won't get corrupted by
 * the loaded 2bl and kernel.
 */

void video_keep(void)
{
    void *saved;
    DWORD phys_saved;
    unsigned long saved_size;
    void *ns;
    DWORD phys_ns;
    DWORD phys_frame_data;

    saved = AvGetSavedDataAddress();
    if (!saved) {
	dprintf(" * No saved data. Turning video output off.\n");
	video_off();
	return;
    }

    saved_size = MmQueryAllocationSize(saved);

    ns = MmAllocateContiguousMemoryEx(saved_size,
				      MIN_SHADOW_ROM, MAX_SHADOW_ROM,
				      0, 0x404);
    if (!ns) {
	dprintf(" * Memory allocation failed for new framebuffer.\n");
	return;
    }

    phys_saved = MmGetPhysicalAddress(saved);
    phys_ns = MmGetPhysicalAddress(ns);

    if ((phys_saved + saved_size) > phys_ns &&
	(phys_ns + saved_size) > phys_saved) {
	/* Overlap. Current frame buffer position good enough. */
	MmFreeContiguousMemory(ns);
	return;
    }

    memcpy(ns, saved, saved_size);

    phys_frame_data = *(DWORD *)(saved + 4) - phys_saved + phys_ns;
    *(DWORD *)(ns + 4) = phys_frame_data;

    __asm__ __volatile__(
	"sfence\n"
	"movl	%%eax,0xfd600800\n"
	"1:jmp	1f\n" /* uh, waits needed? */
	"1:jmp	1f\n"
	"1:jmp	1f\n"
	"1:jmp	1f\n"
	"1:\n"
	"movl	$1,0xfd600100\n"
	: /* no output */
	: "a" (phys_frame_data));

    MmPersistContiguousMemory(ns, saved_size, 1); /* serves waits too =) */

    __asm__ __volatile__(
	"1:\n"
	"testb	$1,0xfd600100\n"
	"jz	1b\n"
	);

    AvSetSavedDataAddress(ns);
    MmFreeContiguousMemory(saved);
}
